from sklearn.metrics import accuracy_score, balanced_accuracy_score
import matplotx
from matplotlib import pyplot as plt


def bacc_acc(predicitions, true_values):
    preds = []
    trues = []
    accs = []
    baccs = []
    for pre, tru in zip(predicitions, true_values):
        preds.append(pre)
        trues.append(tru)
        accs.append(accuracy_score(trues, preds))
        baccs.append(balanced_accuracy_score(trues, preds))
    return accs, baccs


def save_as(filename='iv.png', x_axis=None, xlabel=None, ylabel=None, **kwargs):
    plt.clf()
    for label, values in kwargs.items():
        if x_axis is None:
            x_axis = list(range(len(values)))
        assert len(values) == len(x_axis)  # TODO: Make this error pretty
        plt.plot(x_axis, values, label=label)
    matplotx.line_labels()
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)
    if filename[-4:] != '.png':
        filename = filename + '.png'
    # filename.replace(' ', '_')
    plt.savefig(filename)



import numpy as np
import time

def independent_validation(classifier, X, y, batch_size=1, start_trainset_size=10, scaler=None, talk=False):
    # verdoppeln der batch_size halbiert die nötige Zeit circa.
    start_trainset_size = max(batch_size, start_trainset_size)

    shuffled_indices = np.random.permutation(len(X))
    X = np.array(X)[shuffled_indices]
    y = np.array(y)[shuffled_indices]

    classes = np.unique(y)

    X_train = X[:start_trainset_size]
    y_train = y[:start_trainset_size]
    X = X[start_trainset_size:]
    y = y[start_trainset_size:]
    classifier.fit(X_train, y_train)  # do I need to use partial_fit here for those that can use that?
    true_values = y.tolist()
    predictions = []  # this seems to be faster than using an array.
    t = time.time()
    if scaler is not None:
        scaler.fit(X_train)
    while True:
        if batch_size < len(X):
            X_batch = X[:batch_size]
            y_batch = y[:batch_size]
            X = X[batch_size:]
            y = y[batch_size:]
        else:
            X_batch = X
            y_batch = y

        if scaler is not None:
            preds = classifier.predict(scaler.transform(X_batch))
        else:
            preds = classifier.predict(X_batch)
        predictions.extend(preds.tolist())  #

        X_train = np.vstack([X_train, X_batch])  # wie schnell ist das?
        # Alternative: Einen Datensatz nehmen und jedesmal neu indexieren.
        y_train = np.concatenate([y_train, y_batch])

        if hasattr(classifier, 'partial_fit') and scaler is None:
            classifier.partial_fit(X_batch, y_batch, classes=classes)
        else:
            if scaler is not None:
                scaler.fit(X_train)
                classifier.fit(scaler.transform(X_train), y_train)
            else:
                classifier.fit(X_train, y_train)

        if np.array_equal(X_batch, X):
            return predictions, true_values
        else:
            if talk:
                print(f'Data samples:\t\t{len(predictions)} / {len(true_values)}')
                print(f'estimated time:\t\t {round((time.time() - t))} / {round((time.time() - t) / ((len(predictions) / len(true_values))**2))}')


from plot_beta_file import plot_beta
from independent_validation_file import independent_validation

import matplotlib.pyplot as plt
import scipy.stats as stats
import numpy as np

def plot_multiple(classifiers, X, y, batch_size=1, start_trainset_size=10, scaler=None, talk=False, filename='beta.png', all_betas=False):
    dists = {}
    for clf in classifiers:
        if talk:
            print(f'Starting with classifier {clf.__class__.__name__}')
        predictions, true_values = independent_validation(clf, X, y, batch_size=batch_size, start_trainset_size=start_trainset_size, scaler=scaler, talk=talk)
        mean, var = plot_beta(predictions, true_values, filename=filename, all_betas=all_betas, return_early=True)
        dists.update({clf: (mean, var)})
    plt.clf()
    x = np.linspace(0, 1, 10000)
    for clf in dists.keys():
        mean, var = dists[clf]
        y = stats.norm.pdf(x, loc=mean, scale=math.sqrt(var))
        plt.plot(x, y, label=clf.__class__.__name__)
    plt.savefig(f'{filename}')


import matplotlib.pyplot as plt

import numpy as np
import scipy.stats as stats

import math

def plot_beta(predictions, true_values, filename='beta.png', all_betas=True, return_early=False):
    plt.clf()
    arr = np.array([predictions, true_values])
    cor = arr[0] == arr[1]

    x = np.linspace(0, 1, 10000)

    labels = np.unique(arr[1])
    _ = labels.shape[0]

    means = []
    vars = []

    for i in labels:
        mask = arr[1] == i
        a = cor[mask].sum() + 1
        b = mask.sum() - a + 2  # 2 because the minus a needs to be balanced.

        mean, var, skew, kurt = stats.beta.stats(a, b, moments='mvsk')
        means.append(mean)
        vars.append(var)

        if all_betas:
            plt.plot(x, stats.beta.pdf(x, a, b))

    # approximation as normal distributions to sum up the distributions
    mean = sum(means) / len(means)
    var = sum(vars) / (len(vars) ** 2)
    if return_early:
        return mean, var
    y = stats.norm.pdf(x, loc=mean, scale=math.sqrt(var))
    plt.plot(x, y, label='bacc')
    a = cor.sum() + 1
    b = cor.shape[0] - a + 2  # 2 because the minus a needs to be balanced.
    y = stats.beta.pdf(x, a, b)
    plt.plot(x, y, label='acc')
    plt.savefig(f'{filename}')
    return mean, var  # returning the parameters for the bacc